// Code generated by "stringer -type MapFlags"; DO NOT EDIT.

package sys

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[BPF_F_NO_PREALLOC-1]
	_ = x[BPF_F_NO_COMMON_LRU-2]
	_ = x[BPF_F_NUMA_NODE-4]
	_ = x[BPF_F_RDONLY-8]
	_ = x[BPF_F_WRONLY-16]
	_ = x[BPF_F_STACK_BUILD_ID-32]
	_ = x[BPF_F_ZERO_SEED-64]
	_ = x[BPF_F_RDONLY_PROG-128]
	_ = x[BPF_F_WRONLY_PROG-256]
	_ = x[BPF_F_CLONE-512]
	_ = x[BPF_F_MMAPABLE-1024]
	_ = x[BPF_F_PRESERVE_ELEMS-2048]
	_ = x[BPF_F_INNER_MAP-4096]
}

const _MapFlags_name = "BPF_F_NO_PREALLOCBPF_F_NO_COMMON_LRUBPF_F_NUMA_NODEBPF_F_RDONLYBPF_F_WRONLYBPF_F_STACK_BUILD_IDBPF_F_ZERO_SEEDBPF_F_RDONLY_PROGBPF_F_WRONLY_PROGBPF_F_CLONEBPF_F_MMAPABLEBPF_F_PRESERVE_ELEMSBPF_F_INNER_MAP"

var _MapFlags_map = map[MapFlags]string{
	1:    _MapFlags_name[0:17],
	2:    _MapFlags_name[17:36],
	4:    _MapFlags_name[36:51],
	8:    _MapFlags_name[51:63],
	16:   _MapFlags_name[63:75],
	32:   _MapFlags_name[75:95],
	64:   _MapFlags_name[95:110],
	128:  _MapFlags_name[110:127],
	256:  _MapFlags_name[127:144],
	512:  _MapFlags_name[144:155],
	1024: _MapFlags_name[155:169],
	2048: _MapFlags_name[169:189],
	4096: _MapFlags_name[189:204],
}

func (i MapFlags) String() string {
	if str, ok := _MapFlags_map[i]; ok {
		return str
	}
	return "MapFlags(" + strconv.FormatInt(int64(i), 10) + ")"
}
